#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <memory.h>
#include <string.h>

#include "HashTable.h"

// Hash Function
Index hashFunction(const char *str)
{
	Index idx = 0;
	const char *ptr = NULL;
	int val = 0;
	ptr = str;

	while (*ptr != '\0') {
		int tmp;
		val = (val << 4) + (*ptr);
		if (tmp = (val & 0xf0000000)) {
			val = val ^ (tmp >> 24);
			val = val ^ tmp;
		}
		ptr++;
	}
	idx = val % TABLE_SIZE;

	return idx;
}

// Init the hash table
void initTable(HashTable *table)
{
	int i = 0;
	assert(table != NULL);

	// Set all the entries of the table as empty
	for (i = 0; i < TABLE_SIZE; i++) {
		table->hashTable[i] = NULL;
		table->sizes[i] = 0;
	}
}

// Destroy all the contents of the table
void destroyTable(HashTable * table)
{
	int i = 0;
	for (i = 0; i < TABLE_SIZE; ++i) {
		// Delete all the lists of the hash table
		if (table->hashTable[i] != NULL) {
			HashEntry *node = table->hashTable[i];
			HashEntry *next = node->next;
			while (node != NULL) {
				next = node->next;
				destroyStudent(&node->student);
				free(node);
				node = next;
			}
		}
	}
	
}




// Insert an entry into the hash table
void insertEntry(HashTable *table, Student *student)
{	Index pos = (Index)0;
	HashEntry *entry = NULL;
	
	// Find the position to add the student
	pos = hashFunction(student->id);

	// Allocate data to add in the hash table
	entry = malloc(sizeof(HashEntry));
	copyStudent(student, &entry->student);

	// Add the entry at the beginning of the entry list
	entry->next = table->hashTable[pos];
	table->hashTable[pos] = entry;
	// Hash table entry size
	++table->sizes[pos];

#ifdef _DEBUG
	fprintf(stdout, "Adding student with ID %s in hash table at position %d\n", student->id, pos);
	if (entry->next != NULL) {
		fprintf(stderr, "\tCollision.. Adding student in collision list!\n");
	}

#endif // _DEBUG

}

// Find a student in the hash table
HashEntry *findEntry(HashTable *table, const char *studentID)
{
	Index pos = hashFunction(studentID);
	HashEntry *entry = NULL;

	entry = table->hashTable[pos];
	while (entry != NULL) {
		if (strcmp(entry->student.id, studentID) == 0)
			return entry;
		else
			entry = entry->next;
	}
	return NULL;
}

// Delete an entry from the hash table
void deleteEntry(HashTable *table, const char *studentID)
{
	Index pos = hashFunction(studentID);
	HashEntry *entry = NULL;
	HashEntry *prev = NULL;

	entry = table->hashTable[pos];
	while (entry != NULL) {
		if (strcmp(entry->student.id, studentID) == 0) {
			// Found the entry --> delete it
			if (prev == NULL) {
				table->hashTable[pos] = entry->next;
			}
			else {
				prev->next = entry->next;
			}
			destroyStudent(&entry->student);
			free(entry);
			--table->sizes[pos];
			return;
		}
		else {
			prev = entry;
			entry = entry->next;
		}
	}
#ifdef _DEBUG
	fprintf(stderr, "WARNING: Entry with ID %s not found!\n", studentID);
#endif // _DEBUG
}

